from meiga.error import Error
from meiga.result import Result


class ReturnErrorOnFailure(Error):
    def __init__(self, result: Result):
        Exception.__init__(self)
        self.result = result


def meiga(func):
    def _meiga(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ReturnErrorOnFailure as exc:
            return exc.result

    return _meiga


def return_on_failure(result: Result):
    if result.is_failure:
        raise ReturnErrorOnFailure(result)


def get_value_or_return_on_failure(result: Result):
    if result.is_failure:
        raise ReturnErrorOnFailure(result)
    else:
        return result.value



from typing import TypeVar, Generic

from meiga.no_given_value import NoGivenValue, NoGivenValueClass

TS = TypeVar("TS")  # Success Type
TF = TypeVar("TF")  # Failure Type


class Result(Generic[TS, TF]):
    _is_success: bool = False

    def __init__(self, success: TS = NoGivenValue, failure: TF = NoGivenValue) -> None:
        self._value_success = success
        self._value_failure = failure
        self._assert_values()

    def __repr__(self):
        status = "failure"
        value = self.value.__class__.__name__
        if self._is_success:
            status = "success"
            value = self.value
        return f"Result[status: {status} | value: {value}]"

    def __eq__(self, other):
        if isinstance(other, Result):
            return (
                self._value_success == other._value_success
                and self.value == other.value
            )
        return False

    def _assert_values(self):
        if isinstance(self._value_success, NoGivenValueClass) and isinstance(
            self._value_failure, NoGivenValueClass
        ):
            raise TypeError(
                "Result is a monad, it must be a success or a failure. "
                "Please model your result selecting only one type [success or failure]."
            )
        elif not isinstance(self._value_success, NoGivenValueClass) and not isinstance(
            self._value_failure, NoGivenValueClass
        ):
            raise TypeError(
                "Result is a monad, it cannot be success and failure at the same time. "
                "Please model your result selecting only one type [success or failure]."
            )
        elif not isinstance(self._value_success, NoGivenValueClass):
            self._is_success = True

    def get_value(self):
        if self._is_success:
            return self._value_success
        else:
            return self._value_failure

    def is_success(self):
        return self._is_success

    def is_failure(self):
        return not self._is_success

    value = property(get_value)
    is_success = property(is_success)
    is_failure = property(is_failure)



from typing import Any

from meiga.result import Result


def assert_success(result: Result, value_is_instance_of: Any = None):
    assert (
        result.is_success
    ), f"result is not success as expected. Given failure value is {result.value}"
    if value_is_instance_of:
        assert isinstance(result.value, value_is_instance_of), (
            f"Value is not instance of {value_is_instance_of}."
            f" Given value is {result.value}"
        )



from .meiga_decorator import meiga, return_on_failure, get_value_or_return_on_failure
from .logging_decorator import log_on_start, log_on_end, log_on_error

__all__ = [
    "meiga",
    "return_on_failure",
    "get_value_or_return_on_failure",
    "log_on_start",
    "log_on_end",
    "log_on_error",
]
